// Copyright 2020 The OPA Authors.  All rights reserved.
// Use of this source code is governed by an Apache2
// license that can be found in the LICENSE file.

package ast

const (
	compileStageComprehensionIndexBuild = "compile_stage_comprehension_index_build"
)
